/*
  Copyright (C) 2004-2006  Dmitry V. Levin <ldv@altlinux.org>

  The setproctitle library interface.

  This library is free software; you can redistribute it and/or
  modify it under the terms of the GNU Lesser General Public
  License as published by the Free Software Foundation; either
  version 2.1 of the License, or (at your option) any later version.

  This library is distributed in the hope that it will be useful,
  but WITHOUT ANY WARRANTY; without even the implied warranty of
  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the GNU
  Lesser General Public License for more details.

  You should have received a copy of the GNU Lesser General Public
  License along with this library; if not, write to the Free Software
  Foundation, Inc., 51 Franklin St, Fifth Floor, Boston, MA  02110-1301  USA
*/

/* http://sisyphus.ru/srpm/Sisyphus/setproctitle/ */

#include <assert.h>
#include <errno.h>
#include <stdarg.h>
#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <unistd.h>

#include "setproctitle.h"

int _init(int argc, char *argv[], char *envp[]);

static char *title_buffer = 0;
static size_t title_buffer_size = 0;
static char *title_progname, *title_progname_full;

int
setproctitle_progname(const char *name)
{
	/* check arguments */
	if(!name)
	{
		errno = EINVAL;
		return -1;
	}

	/* strdup() leaks memory if this function is used more than
	   once, but usually this function is only used once after the
	   program start to set a custom name. */
	char *s=strdup(name);
	if(!s)
	{
		errno = ENOMEM;
		return -1;
	}

	/* we could alloc new memory in strdup(), so we can assign to
	   title_progname now */
	title_progname=s;

	return 0;
}

int
setproctitle(const char *fmt, ...)
{
	/* check arguments */
	if (!title_buffer || !title_buffer_size)
	{
		errno = ENOMEM;
		return -1;
	}

	/* we calculate the length of the program name in written */
	ssize_t written=0;

	if (fmt)
	{
		ssize_t written2;
		va_list ap;

		/* should we omit the program name? */
		if(fmt[0] == '-')
			++fmt;	/* skip over '-' */
		else
		{
			/* no, print program name to title_buffer */
			written = snprintf(title_buffer, title_buffer_size, "%s: ", title_progname);
			if (written < 0 || (size_t) written >= title_buffer_size)
				return -1;
		}

		/* print arguments to title_buffer */
		va_start(ap, fmt);
		written2 = vsnprintf(title_buffer + written, title_buffer_size - written, fmt, ap);
		va_end(ap);
		if (written2 < 0 || (size_t) written2 >= title_buffer_size - written)
			return -1;

		/* adjust title length */
		written+=written2;
	}
	else
	{
		/* no arguments, just print program name */
		written = snprintf(title_buffer, title_buffer_size, "%s", title_progname);
		if (written < 0 || (size_t) written >= title_buffer_size)
			return -1;
	}

	/* check we have calculated written correctly */
	assert(written == (int)strlen(title_buffer));

	/* fill the rest of the title_buffer with 0 */
	memset(title_buffer + written, '\0', title_buffer_size - written);

	return 0;
}

int
_init(int argc, char *argv[], char *envp[])
{
	char   *begin_of_buffer = 0, *end_of_buffer = 0;
	int     i;

	for (i = 0; i < argc; ++i)
	{
		if (!begin_of_buffer)
			begin_of_buffer = argv[i];
		if (!end_of_buffer || end_of_buffer + 1 == argv[i])
			end_of_buffer = argv[i] + strlen(argv[i]);
	}

	for (i = 0; envp[i]; ++i)
	{
		if (!begin_of_buffer)
			begin_of_buffer = envp[i];
		if (!end_of_buffer || end_of_buffer + 1 == envp[i])
			end_of_buffer = envp[i] + strlen(envp[i]);
	}

	if (!end_of_buffer)
		return 0;

	char **new_environ = (char**) malloc((i + 1) * sizeof(envp[0]));

	if (!new_environ)
		return 0;

	for (i = 0; envp[i]; ++i)
		if (!(new_environ[i] = strdup(envp[i])))
			goto cleanup_enomem;
	new_environ[i] = 0;

	if (program_invocation_name)
	{
		title_progname_full = strdup(program_invocation_name);

		if (!title_progname_full)
			goto cleanup_enomem;

		char   *p = strrchr(title_progname_full, '/');

		if (p)
			title_progname = p + 1;
		else
			title_progname = title_progname_full;

		program_invocation_name = title_progname_full;
		program_invocation_short_name = title_progname;
	}

	environ = new_environ;
	title_buffer = begin_of_buffer;
	title_buffer_size = end_of_buffer - begin_of_buffer;

	return 0;

      cleanup_enomem:
	for (--i; i >= 0; --i)
		free(new_environ[i]);
	free(new_environ);
	return 0;
}
